/*###########################################################################
#                                                                           #
#                             fimage v1.0                                   #
#        Copyright (C) 2001 2002  David TAILLANDIER - SOLEIL VERT           #
#                                                                           #
#                                                                           #
#############################################################################

fimage is a DOS program to create/use 1.44 Mb floppy diskette images.

Full program into one source file.


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.



Revision history :

jully 12th 2001  v0.8  David TAILLANDIER, SOLEIL VERT. Initial release. Code
   initialy taken from 'rawrite' but finaly rewriten from scratch with
   some idea from Michel CARAMEZ, SOLEIL VERT.
   From rawrite comes the knowledge of 64kb boundary DMA problem.

jully 13th 2001  v0.9  David TAILLANDIER, SOLEIL VERT. Stupid bug fix in
   command-line parsing. Tracking for a bug in Windows DOS emulation mode.

february 28th 2002 v1.0  David TAILLANDIER. Removed support for diskette
   size other than 1.44 Mb   (former versions had support for 360, 1.2 and
   720 too).

*/

/*############################# includes ##################################*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>


/*############################## defines & Co #############################*/

#define  READ    2      /* read diskette TO image    */
#define  WRITE   3      /* write diskette FROM image */

#define  BUFFERSIZE  (512*18)    /* size of one floppy track (1 side) */


/*########################## functions declaration ########################*/

void help_message( void );
void license_message( void );
void error( const char * errortext, ... );
void read_cmd_line( int argc, char * argv[] );
int reset_floppy( int drive );
int rw_floppy( char * buffer, int operation, int drive, int track, int head );


/*############################## global variables #########################*/

int readwrite;
FILE * imagefile;
int drive;
char * buffer;


/*################################# main ##################################*/
int main( int    argc,
          char * argv[]
        )
{
     int track, head;

     printf( "fimage for DOS     v1.0     floppy disk image utility\n"
             "Copyright (C) 2001 2002 David TAILLANDIER - SOLEIL VERT\n"
             "       fimage -h   for help\n"
             "       fimage -l   for license information\n"
             "\n"
           );

     /* allocate track-buffer without 64kb boundary because */
     /* some (old ?) computers have DMA problem             */
     buffer = (char *)malloc( BUFFERSIZE * 2 );   /* *2 to be able to avoid 64kb boundary */
     if( buffer == NULL )
          error( "Unable to allocate %i kb into memory\n", BUFFERSIZE/1024 );
     if(  ( FP_SEG( buffer               ) & 0xf000 ) !=
          ( FP_SEG( buffer+BUFFERSIZE+16 ) & 0xf000 )  )
         buffer += 16 + BUFFERSIZE;   /* use second half of double-sized buffer */

     read_cmd_line( argc, argv );

     /* check image size against floppy size */
     if ( readwrite == WRITE )
       {
          if ( fseek(imagefile,0,SEEK_END) != 0 )
               error( "Problem reading image file\n" );
          if ( ftell(imagefile) != 1474560L )
               error( "Image file must have a size of 1.44 Mb\n" );
          if ( fseek(imagefile,0,SEEK_SET) != 0 )
               error( "Problem reading image file\n" );
       }

     /* the main thing */
     if ( readwrite == READ )
          printf( "Reading floppy" );
     else
          printf( "Writing floppy" );
     delay( 300 );  /* because Windows 9x doesn't display    */
                    /* correcly while accessing floppy drive */

     for ( track=0; track!=80; track++ )
       {
          for ( head=0; head!=2; head++ )
            {
               if ( readwrite == READ )
                 {
                    if ( rw_floppy(buffer,READ,drive,track,head) != 0 )
                         error( "\nFloppy diskette read error\n" );
                    if ( fwrite(buffer,1,512*18,imagefile) != 512*18 )
                         error( "\nError when writing to image file\n" );
                 }
               else
                 {
                    if ( fread(buffer,1,512*18,imagefile) != 512*18 )
                         error( "\nError when reading image file\n" );
                    if ( rw_floppy(buffer,WRITE,drive,track,head) != 0 )
                         error( "\nFloppy diskette write error\n" );
                 }
            }   /* for( head ) */
       }   /* for( track ) */

     fclose( imagefile );
     printf( "\nDone.\n" );
     return 0;
}

/*########################### read_cmd_line ###############################*/
void read_cmd_line( int    argc,
                    char * argv[]
                  )
{
     int arg_num;
     char * filenameptr;
     int nowait;
     char * errortext;
     char filename[200];

     arg_num = 0;
     nowait = 0;
     drive = -1;
     filenameptr = NULL;
     errortext = "Command line error\n";

     /* read command line */
     while( arg_num+1 < argc )
       {
          arg_num++;
          if ( argv[arg_num][0] != '-' )    /* filename */
            {
                     if ( filenameptr != NULL )
                        error( "%s", errortext );
                     filenameptr = argv[arg_num];
                     readwrite = WRITE;
                     continue;
            }

          switch( argv[arg_num][1] )
            {
               case 'w' :   /* write floppy */
               case 'r' :   /* read floppy  */
                            if ( filenameptr != NULL )
                                 error( "%s" "'-r' and -'w' and file name can only be specified once\n", errortext );
                            if ( argv[arg_num][1] == 'w' )
                                 readwrite = WRITE;
                            else
                                 readwrite = READ;
                            arg_num++;
                            if ( arg_num+1 > argc )
	                         error( "%s" "File name expected after -w or -r option\n", errortext );
                            filenameptr = argv[arg_num];
                            break;  /* the final work will be done on the next iteration */

               case 'd' :   /* drive spec */
                            if ( drive != -1 )
                                 error( "%s" "-d option can only be specified once\n", errortext );
                            arg_num++;
                            if ( arg_num > argc )
                                 error( "%s" "Drive letter expected after -d option\n", errortext );
                            drive = argv[arg_num][0];
                            if ( argv[arg_num][1] != 0 )
                                 error( "%s" "'A' or 'B' expected after '-d' option\n", errortext );
                            break;

               case 'h' :   /* help */
                            help_message();

               case 'l' :   /* license */
                            license_message();

               case 'n' :   /* no wait */
                            nowait = 1;
                            break;

               default :   /* ahhh... well... please reboot user */
                           error( "%s", errortext );
            }   /* switch */
       }   /* while */

     /* verify drive name */
     if ( drive == -1 )
          drive = 0;      /* A */
     else
       {
          drive -= 'A';
          if ( drive > 26 )
               drive -= 'a'-'A';
          if ( drive!=0 && drive!=1 )
               error( "%s" "drive name must be A or B\n", errortext );
       }

     /* ask for file name and read/write */
     if ( filenameptr == NULL )
       {
          printf("Do you want to read (R) or to write (W) diskette ?  ");
          fgets( filename, 5, stdin );
          switch( filename[0] )
            {
               case 'r' :
               case 'R' :   readwrite = READ;
                            break;
               case 'w' :
               case 'W' :   readwrite = WRITE;
                            break;

               default :    error( "\nYou must type 'R' or 'W'\n" );
            }
 
          printf( "Enter file name : " );
          fgets( filename, 200, stdin );
          *strchr( filename, '\n' ) = 0;
          filenameptr = filename;
       }

     /* verify file name */
     if ( readwrite == READ )
          imagefile = fopen( filenameptr, "wb" );
     else
          imagefile = fopen( filenameptr, "rb" );
     if ( imagefile == NULL )
          error( "Unable to access to specified image file\n" );

     /* wait for diskette */
     if ( nowait == 0 )
       {
          printf("\n");
          printf("Insert diskette and press [ENTER]  ");
          fgets( "      ", 3, stdin );
       }
}

/*############################# reset_floppy ##############################*/
/* return non-zero on error */
int reset_floppy( int drive )
{
     union REGS regs;

     regs.h.ah = 0;
     regs.h.dl = (unsigned char)drive;
     int86( 0x13, &regs, &regs );
     return( regs.h.ah );
}

/*################################ rw_floppy ##############################*/
/* return non-zero on error */
int rw_floppy( char * buffer,
               int    operation,   /* 2=read  3=write */
               int    drive,
               int    track,
               int    head
             )
{
     union REGS regs;
     struct SREGS sregs;
     int i;

     for( i=0; i!=2; i++ )         /* try 2 times */
       {
          regs.h.ah = operation;
          regs.h.al = 18;
          regs.h.ch = track;
          regs.h.cl = 1;
          regs.h.dh = head;
          regs.h.dl = drive;
          sregs.es  = FP_SEG( buffer );
          regs.x.bx = FP_OFF( buffer );
          int86x( 0x13, &regs, &regs, &sregs );

          if( regs.h.ah == 0 )
               break;
          reset_floppy( drive );
       }

     return( regs.h.ah );
}

/*############################### help_message ############################*/
void help_message( void )
{
     error( "\n"
            "This program creates a file which is an image of a floppy diskette.\n"
            "You can then store floppy images, transmit them by internet, and so on.\n"
            "This program can also restore a floppy from an image. Thoses images are\n"
            "'standard' (raw copy) and can be handled by other programs like 'rawrite'\n"
            "or 'dd' unix command and a lot of other utilities.\n"
            "\n"
            "options are :\n"
            "\n"
            "   -w <filename>       write floppy (from image file)\n"
            "or <filename>          write floppy (from image file)\n"
            "   -r <filename>       read floppy (to image file)\n"
            "   -d <drive-letter>   floppy drive letter (A or B). Default is A\n"
            "   -n                  (no-wait) don't wait for user to insert floppy\n"
            "   -h                  (help) display this help message and exit\n"
            "   -l                  (license) display license information and exit\n"
            "\n"
            "\n"
          );

}

/*############################ license_message ############################*/
void license_message( void )
{
     error( "\n"
            "Copyright (C) 2001 2002 David TAILLANDIER - SOLEIL VERT\n"
            "\n"
            "This program is free software; you can redistribute it and/or\n"
            "modify it under the terms of the GNU General Public License\n"
            "as published by the Free Software Foundation; either version 2\n"
            "of the License, or (at your option) any later version.\n"
            "\n"
            "This program is distributed in the hope that it will be useful,\n"
            "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
            "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
            "GNU General Public License for more details.\n"
            "\n"
            "You should have received a copy of the GNU General Public License\n"
            "along with this program; if not, write to the Free Software\n"
            "Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
            "\n"
          );
}

/*############################## error ####################################*/
void error( const char * errortext,
            ...
          )
{
     va_list vargs;

     va_start( vargs, errortext );
     vfprintf( stderr, errortext, vargs );
     va_end( vargs );

     exit( -1 );
}

/*############################# END OF FILE ###############################*/

